function attractor(x)
    local x = x / 300
    return 10 * math.exp(-x*x)
end

function repulsor(x)
    local x = x / 300
    if x > 2 then return 0 end
    return - 1 / (x + 1e-10)
end

function swarm(x)
    local x = x / 100
    if x > 3 then return 0 end
    return 9 * math.exp(-x*x) - (1 / (x + 1e-10))
end

BOIDS = {}
BOIDCOUNT = 60
ATTRACTORS = {}
REPULSORS = {}
WIDTH = 800
HEIGHT = 600
MAX_SPEED = 50
SIMULATION_SPEED = 3

---------------------------------------------------

vector = {
    new = function(x,y)
        local x = x or 0
        local y = y or 0
        local v = {x=x,y=y}
        setmetatable(v, vector)
        return v
    end,

    dist = function(a,b)
        return (a-b):len()
    end,
    __add = function(a, b)
        return vector.new(a.x + b.x, a.y + b.y)
    end,
    __sub = function(a, b)
        return vector.new(a.x - b.x, a.y - b.y)
    end,
    __mul = function(a, b)
        if type(a) == "number" then
            return vector.new(a * b.x, a * b.y) end
        return vector.new(b * a.x, b * a.y)
    end,
    __div = function(a,b)
        return a * (1 / b)
    end,
    len2 = function(a)
        return a.x * a.x + a.y * a.y
    end,
    len = function(a)
        return math.sqrt(a:len2())
    end,
    normalized = function(a)
        local l = a:len()
        if l > 0 then return a / a:len() end
        return vector.new(0,0)
    end
}
vector.__index = vector

boid = {
    new = function(mass, leader, x, y)
        local x = x or math.random(0, WIDTH)
        local y = y or math.random(0, HEIGHT)
        local b = {
            leader = leader or false,
            pos = vector.new(x,y), 
            vel = vector.new(0,0),
            force = vector.new(0,0),
            mass = mass}
        setmetatable(b, boid)
        return b
    end,

    draw = function(self)
        -- draw boid
        if self.leader then
            love.graphics.setColor(0,255,0,255)
        else
            love.graphics.setColor(0,0,0,255)
        end
        love.graphics.circle('fill', self.pos.x, self.pos.y, 5, 3)
--        if self.leader then
--            love.graphics.setLine(1)
--            love.graphics.setColor(0,0,0,40)
--            love.graphics.line(self.pos.x, self.pos.y, love.mouse.getX(), love.mouse.getY())
--        end
        -- draw velocity
        love.graphics.setLine(1)
        love.graphics.setColor(0,100,255,100)
        love.graphics.line(self.pos.x, self.pos.y,
            self.pos.x + self.vel.x, self.pos.y + self.vel.y)
        -- draw force
--        love.graphics.setLine(5)
--        love.graphics.setColor(255,128,0,100)
--        love.graphics.line(self.pos.x, self.pos.y,
--            self.pos.x + self.force.x, self.pos.y + self.force.y)
    end,
    move = function(self, dt)
        self.vel = self.vel + self.force / self.mass * dt
        if self.vel:len() > MAX_SPEED then
            self.vel = self.vel:normalized() * MAX_SPEED
        end
        self.pos = self.pos + self.vel * dt
        if self.pos.x < 0 then
            self.pos.x = 0 end
        if self.pos.x >= WIDTH then
            self.pos.x = WIDTH-1 end
        if self.pos.y < 0 then
            self.pos.y = 0 end
        if self.pos.y >= HEIGHT then
            self.pos.y = HEIGHT-1 end
    end
}
boid.__index = boid

function update_force(b, arr, accessor, dfunc)
    local dist, direction, pos
    for i = 1,#arr do
        pos = accessor(arr[i])
        if pos ~= b.pos then
            direction = pos - b.pos
            dist = direction:len2()
            if dist < 300 * 300 then
                dist = math.sqrt(dist)
                if dist == 0 then dist = 1e-10 end
                direction = direction:normalized() * dfunc(dist)
                b.force = b.force + direction
            end
        end
    end
end

function get_boids_force()
    for _, b in pairs(BOIDS) do
        if b.leader then
--            local dmouse = vector.new(love.mouse.getX(), love.mouse.getY()) - b.pos
--            dmouse = dmouse:normalized() * 2
--            b.force = dmouse
            b.force = vector.new(0,0)
            update_force(b, ATTRACTORS, function(v) return v end, attractor)
            update_force(b, REPULSORS, function(v) return v end, repulsor)
        else
            b.force = vector.new(0,0)
        end
        update_force(b, BOIDS, function(b) return b.pos end, swarm)
        -- dont come to close to the walls
        walls = {repulsor(b.pos.x), repulsor(WIDTH - b.pos.x), -- left, right
                 repulsor(b.pos.y), repulsor(HEIGHT - b.pos.y)} -- rop, bottom
        b.force = b.force + (vector.new(-1,  0) * walls[1] 
                          + vector.new( 1,  0) * walls[2]
                          + vector.new( 0, -1) * walls[3]
                          + vector.new( 0,  1) * walls[4]) * .8
    end
end

function love.load()
    math.randomseed( os.time() )
    love.graphics.setBackgroundColor(255,255,255)
    for i = 1,BOIDCOUNT do
        BOIDS[#BOIDS + 1] = boid.new(1 + 3 * math.random(), math.random() > .8)
    end

    -- this has to be called twice so the mouse gets the update ... strange
    love.graphics.setMode(WIDTH, HEIGHT)
    love.graphics.setMode(WIDTH, HEIGHT)

end

function love.update(dt)
    get_boids_force()
    for _,b in pairs(BOIDS) do
        b:move(dt * SIMULATION_SPEED)
    end
end

function love.draw()
	-- draw attractors
	love.graphics.setColor(0,255,0,128)
	for _, a in pairs(ATTRACTORS) do
		love.graphics.circle('fill', a.x, a.y, 10)
	end

	-- draw repulsors
	love.graphics.setColor(255,0,0,128)
	for _, a in pairs(REPULSORS) do
		love.graphics.circle('fill', a.x, a.y, 10)
	end

	-- draw boids
	for _,b in pairs(BOIDS) do
		b:draw()
	end

	local timer = love.timer
	love.graphics.setColor(0,0,0,255)
	love.graphics.print(string.format("FPS: %d, boids: %d", timer.getFPS(), #BOIDS), 2, 10)
	love.graphics.print(string.format("Mouse: %dx%d", love.mouse.getX(), love.mouse.getY()), 2, 20)
end

function love.mousereleased(x,y, btn)
    if btn == 'l' then
        ATTRACTORS[#ATTRACTORS + 1] = vector.new(x, y)
    elseif btn == 'r' then
        REPULSORS[#REPULSORS + 1] = vector.new(x, y)
    else
        BOIDS[#BOIDS+1] = boid.new(1 + 3 * math.random(), math.random() > .8, x, y)
    end
end
